//
// HPDisk - Emulate a HP9121 diskette drive, using an SD card
//
// 20131103 ag	0.1	First working version
// 20131120 ag	0.2	Rearranged pins so that we can interrupt on change
// 20131124 ag	0.3	Changed so that we now write 256-byte sectors, two per real sector. Thus the resulting file is a true image of the LIF "disk"
// 20131125 ag	0.4 Small tweaks. Does now work properly with SDEBUG undefined. Needed a small delay in the loop
// 20131208 ag	0.5 Corrected the reported status for write protect, added handling of Amigo Wakeup, added support for hardware PP
// 20140108 ag	0.6 Enabled support for FAT32
// 20140113 ag	0.7 Will now flash an errorcode on the leds if it cannot open the file on the SD-card: Se the FRESULT enum in pff.h
// 20140120 ag	0.8 Added support for SS/80, selected through #define for now
// 20140123 ag	0.9 Shortened the GPIB timeout, waiting for DAV and added an extra delay in case GPIB Receive returns timeout
// 20141025 ag	0.10 Check IFC in all important loops and resets on IFC.
//					 Changed GPIB timeout so that we catch slow IDENT
//					 Changed write logic so that we now handle writes longer than 256 bytes
//					 GPIB receive buffer is now 256 bytes - simplifies things
//					 Added support for command 0x37 - Initialise and 0x34 No Op
//					 Honours the given Length on both read and writes
// 20140129 ag 	0.11 Now does writes in 256-byte chunks also. Fixed the extra character output in DumpData
// 20140131 ag	0.12 Fixed the problem where we would sen EOI far to early on > 256 byte sends
// 20140202 ag	0.13 Do bounds checking on seeks. return at least one byte tagged with EOI on failed reads
//					 Do a complete reint in case we suddently get 5 - not open, on file ops
// 20140204 ag	0.14 Supports both Amigo and SS/80 with the same firmware
//					 Reads settings from hpdisk.cfg on SD-card
// 20140209 ag	0.15 We can skip GPIB_TX_CHECK and go directly to PUT_DATA, that saves one loop through the state-machine
//					 We can probably skip the delay in wait to settle the line as a loop takes approx 17us
//					 Still timing problems if we make it too fast by eliminating all debug sprintf's
//
// Comments on bugs and suggested improvements are welcome. Mail me at gpib@dalton.ax
//
// Project page: http://www.dalton.ax/hpdisk
//
// Todo:


#include <p18f4620.h>
#include <delays.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "gpib.h"
#include "amigo.h"
#include "ss80.h"
#include "sio.h"


void format_disk(void);
void write_sectors(void);
void DumpData(unsigned char *Ptr,int length);
int SD_mount(void);
void Error_blink(int code);
int Read_config(void);
void Time(void);

#pragma config WDT = OFF        //Watchdog timer OFF
#pragma config PBADEN = OFF 	//PORTB 0 to 4 digital - not analogue
#pragma config OSC = HSPLL      //HSPLL oscillator, 8MHz crystal with PLL is 32MHz clock
#pragma config LVP = OFF        //Single supply ICSP OFF
#pragma config FCMEN = OFF      //Fail safe clock monitor OFF
#pragma config IESO = OFF       //
#pragma config PWRT = OFF
#pragma config BOREN = OFF
#pragma config MCLRE = ON
#pragma config STVREN = ON		// Stack full/underflow will cause Reset  


#include "Globals.h"
#include "SD.h"
#include "pff.h"
//
// Initialized data
//
#pragma idata myidata
char file[] = "/lifdata.bin"; 		// Name of LIF image file
char cfgfile[] = "/hpdisk.cfg"; 	// Name of config file

// Uninitialized data
//
#pragma udata
FATFS fs; // File system object
FRESULT rc;

extern char string[MAX_DBG_STRING];
extern unsigned char edge[5];

unsigned char MMC; 				// Memory card status
DWORD fpos;						// File position

#pragma udata my_section_2
char buffer[256];
//
// You need to pre-create the LIF image file, liftata.bin. Use this from DOS for example
//
// echo "This is just a sample line appended to create a big file " > lifdata.bin
//for /L %i in (1,1,14) do type lifdata.bin >> lifdata.bin
//
void main(void)
{
    char Key = 0;
	unsigned int i = 0;
	unsigned char c = 0;
	int nbytes, rc2;
	unsigned char status;
	//
	protocol = 1;				// Default protocol: SS/80
	debuglevel = 0;				// Default loglevel
	//
	LED1_TRIS = 0;
	LED2_TRIS = 0;
	//
	LED1_PORT = 1;
	//
	// Set up Timer0 to count
	//
	T0CON = 0b10000111;
	INTCONbits.TMR0IF = 0;
	//
	sio_setup();			// Init serial port for debugging
	//
	sprintf(string,(const far rom char*) "\r\n\r\nHPDisk v 0.15");
	sio_puts(string);
	sprintf(string,(const far rom char*) "\r\n%S %S", __DATE__,__TIME__);
	sio_puts(string);
	sprintf(string,(const far rom char*) "\r\n\SD SPI DIV: %d", SSPDIV);
	sio_puts(string);
	sprintf(string,(const far rom char*) "\r\n\STKPTR: %X", STKPTR);
	sio_puts(string);
	//
	edge[0]=0;
	edge[1]=0;
	edge[2]=0;
	edge[3]=0;
    Globals_IO_init();  	// Set in and outputs for SD-Card


//SS80_init();
//SS80_Test();
	//
	//
	// Hardware or software Parallell Poll
	//
#if defined(SOFTWARE_PP)
	sprintf(string,(const far rom char*) "\r\nSoftware PP\r\n");
#else
	sprintf(string,(const far rom char*) "\r\nHardware PP\r\n");
#endif
	sio_puts(string);
	delay_ms(500);
	//sprintf(string,(const far rom char*)"\r\nStart %d %d", CD_PORT, WP_PORT);
	//sio_puts(string);
	//
	LED2_PORT = 1; 
	//format_disk();
	rc=SD_mount();
	sprintf(string,(const far rom char*)"\r\n\r\nSD_mount returns %d", rc);
	sio_puts(string);
	if(rc != 0)
	{
		Error_blink(rc);
		while(1);
	}
	//
	Read_config();
	if(protocol)			// SS/80
	{
							//
							// SS/80 drives
							//
#if defined(HP9122D)
		sprintf(string,(const far rom char*) "\r\nSS/80 9122D\r\n");
#endif
#if defined(HP9134L)
		sprintf(string,(const far rom char*) "\r\nSS/80 9134L\r\n");
#endif
		idbyte1 = SS80ID1;
		idbyte2 = SS80ID2;
	}
	else					// Amigo
	{
							// Amigo drives
#if defined(HP9121D)
		sprintf(string,(const far rom char*) "\r\nAmigo 9121D\r\n");
#endif
		idbyte1 = AMIGOID1;
		idbyte2 = AMIGOID2;
	}
	sio_puts(string);
	//
	status = 0;
	//
	gpib_init();			// Init GPIB
	//
	while(1)				// Main loop, forever
	{

		//
    	MMC = pf_open(file); //Open the file
		rc = (int) MMC;
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nOpen %s returns %X", file, MMC);
		sio_puts(string);
#endif
		if(rc != 0)
		{
			Error_blink(rc);
			rc=SD_mount();
			continue;
		}
		//
		if(WP_PORT)
		{
			sprintf(string,(const far rom char*)"\r\n\Card is write protected");
			sio_puts(string);
		}
		//
		// Strategy: Scan GPIB continously. readgpib() returns zero if nothing or a command byte, corresponding to an Amigo Command
		// While reading, readgpib() pushes every byte to the Amigo-routines, a valid command breaks out of the receive loop
		// All command handling is done in the Amigo-routines, with two exceptions: Ident and parallell poll
		//
		amigo_init();
		SS80_init();
		rc2 = 0;
		while(CD_PORT == 0 && IFC_PORT == 1 && rc2 != 5) // SDCard CD is low when card present, so loop here and fall through if card removed (PORTCbits.RC1)
		{
			gpib_ptr = 0;
			
			LED2_PORT = !LED2_PORT;
			
			status = readgpib();
			//if(status == 255) delay_ms(20);
			if((status && status != 255) || gpib_ptr>0)
			{

				sprintf(string,(const far rom char*)" - Status: %d  \r\n", status);

#if defined(SDEBUG)
				sio_puts(string);
				DumpData(gpib_buff, gpib_ptr);
#else
				delay_ms(1);			// Found out the hard way... We really need a short delay here (disks are slow)
#endif
				Time();
			}
			rc = 0;
			if(protocol)				// SS/80
			{

				if(status == SS80_CMD_CLEAR) SS80_cmd_clear();
				if(status == SS80_CMD_DESCRIBE) SS80_cmd_describe();
				if(status == SS80_CMD_SEND_STATUS) SS80_cmd_send_status();
				if(status == SS80_CMD_BUFFERED_READ) rc2 = SS80_cmd_buffered_read();
				if(status == SS80_CMD_BUFFERED_WRITE) rc2 = SS80_cmd_buffered_write();
			}
			else						// Amigo
			{
				
				if(status == AMIGO_CMD_CLEAR) amigo_cmd_clear();
				if(status == AMIGO_CMD_SEND_STATUS) amigo_cmd_send_status();
				if(status == AMIGO_CMD_SEEK) amigo_cmd_seek();
				if(status == AMIGO_CMD_BUFFERED_READ) amigo_cmd_buffered_read();
				if(status == AMIGO_CMD_DSJ) amigo_cmd_dsj();
				if(status == AMIGO_CMD_BUFFERED_WRITE) amigo_cmd_buffered_write();
				if(status == AMIGO_CMD_WAKEUP) amigo_cmd_wakeup();
			}
			if(edge[0]!=0)
			{
				sprintf(string,(const far rom char*)"\r\nEdge: %s  %d", edge, edge[0]);
				sio_puts(string);
			}	
		}
		if(IFC_PORT == 0)
		{
			sprintf(string,(const far rom char*)"\r\n\r\nInterface clear");
			while(IFC_PORT == 0)
				;
		}
		else sprintf(string,(const far rom char*)"\r\n\r\nOops. No card");
		sio_puts(string);
		LED1_PORT = 0; 
		delay_ms(1000);
		LED1_PORT = 1; 
		delay_ms(1000);
	}

}
////
//// For testing. Formats the file with valid LIF-contents
////
//void write_sectors(void)
//{
//DWORD fpos, sector;
//int nbytes, temp,i;
//
//FRESULT rc;
//	for(sector=0;sector < (16L * 2L );sector++) // * 66L
//	{
//		
//		//sprintf(string,(const far rom char*)"\r\nWrite sector: %ld %X ", sector, fpos);
//		//sio_puts((unsigned char*)string);
//		if(sector % 2 == 0)	// even
//		{
//			// Read the upper part of the 512-byte sector, ie sector+1
//			fpos = (sector / 2) * 512L + 256L;
//			rc=pf_lseek(fpos);
//			rc=pf_read(buffer,256,&nbytes);
//			sprintf(string,(const far rom char*)"\r\nRead odd sector: %ld %lX %d", sector, fpos, rc);
//			sio_puts((unsigned char*)string);
//			// Seek to start of sector
//			fpos = sector / 2 * 512L;
//			rc=pf_lseek(fpos);
//			sprintf(string,(const far rom char*)"\r\nWrite even sector: %ld %lX %d", sector, fpos, rc);
//			sio_puts((unsigned char*)string);
//			// Write data
//			temp = strlen(string);
//			rc=pf_write(string, temp, &nbytes);
//			for(i=0;i<256-temp;i++)
//				rc=pf_write(&string[8], 1, &nbytes);
//			// Write upper part
//			rc=pf_write(buffer, 256, &nbytes);
//			// finalise
//			rc =pf_write(0, 0, &nbytes);
//		}
//		else			// odd
//		{
//
//			// Read the lower part of the 512-byte sector, ie sector
//			fpos = sector / 2 * 512L;
//			rc=pf_lseek(fpos);
//			rc=pf_read(buffer,256,&nbytes);
//			sprintf(string,(const far rom char*)"\r\nRead even sector: %ld %lX %d", sector, fpos, rc);
//			sio_puts((unsigned char*)string);
//			// Write lower part
//			rc=pf_write(buffer, 256, &nbytes);
//
//			fpos = (sector / 2) * 512L + 256L;
//			sprintf(string,(const far rom char*)"\r\nWrite odd sector: %ld %lX ", sector, fpos);
//			sio_puts((unsigned char*)string);
//			// Write data
//			temp = strlen(string);
//			rc=pf_write(string, temp, &nbytes);
//			for(i=0;i<256-temp;i++)
//				rc=pf_write(string[8], 1, &nbytes);
//
//			// finalise
//			rc =pf_write(0, 0, &nbytes);
//		}
//	}
//}
//void format_disk(void)
//{
//int nbytes;
//	rc=0;
//	sprintf(string,(const far rom char*)"\r\n\r\nMounting");
//	sio_puts(string);
//	delay_ms(500);
//    rc=pf_mount(&fs); //Mount volume
//	sprintf(string,(const far rom char*)"\r\nMount returns %X", rc);
//	sio_puts(string);
//    MMC = pf_open(file); //Open the file
//	sprintf(string,(const far rom char*)"\r\nOpen returns %X", MMC);
//	sio_puts(string);
//	sprintf(string,(const far rom char*)"\r\nReading");
//	sio_puts(string);
//	memset(buffer,0,10);
//    rc=pf_read(buffer,10,&nbytes);
//	sprintf(string,(const far rom char*)"\r\nRead returns %X, %d bytes read: %s", rc, nbytes, buffer);
//	sio_puts(string);
//	//
//	fpos = 0;
//	memset(buffer,' ',511);
//	memcpy(buffer,&vl,sizeof(vl));	// Volume 
//	rc=pf_lseek(fpos);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nSeek fail %d", rc);
//		sio_puts(string);
//	}
//	rc=pf_write(buffer, 512, &nbytes);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
//		sio_puts(string);
//	}
//	pf_write(0, 0, &nbytes);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
//		sio_puts(string);
//	}
//	fpos = (DWORD)2*512L;
//	memset(buffer,' ',511);
//	de.filetype = 0x100;
//	memcpy(buffer,&de,sizeof(de));
//	memcpy(buffer+sizeof(de),&de,sizeof(de));
//	de.filetype = 0xffff;
//	memcpy(buffer+sizeof(de)+sizeof(de),&de,sizeof(de));
//	rc=pf_lseek(fpos);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nSeek fail %d", rc);
//		sio_puts(string);
//	}
//	rc=pf_write(buffer, 512, &nbytes);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
//		sio_puts(string);
//	}
//	pf_write(0, 0, &nbytes);
//	if(rc)
//	{
//		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
//		sio_puts(string);
//	}
//	rc=pf_mount(0); //unmount volume
//	sprintf(string,(const far rom char*)"\r\nDone");
//	sio_puts(string);
//    while (1)
//	;
//}
//
// Dump an data-area in human-readable form
//
void DumpData(unsigned char *Ptr,int length){
int i,j;
char ch;
	sprintf(string,(const far rom char*)"\r\nDump %X %X\r\n",Ptr, length);
	sio_puts(string);
	for(j=0;j<80&&(j*16<length);j++){
		sprintf(string,(const far rom char*)"\r\n");
		sio_puts(string);
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				ch = *(Ptr+i+j*16);
				sprintf(string, (const far rom char*)" %02X",ch&0xFF);
				sio_puts(string);
			}
		sprintf(string, (const far rom char*)" | ");
		sio_puts(string);
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				if(*(Ptr+i+j*16)>' ') sprintf(string, (const far rom char*)"%c",(*(Ptr+i+j*16))&0xFF); 
				else sprintf(string, (const far rom char*)".");
				sio_puts(string);
			}
	}
	sprintf(string, (const far rom char*)"\r\n");
	sio_puts(string);
}
//
// Prepare the SD card and file for access. Do max 5 retries 
// Returns 0 on success or the actual state if failed
//
int SD_mount()
{
int retries;

	retries = 5;
	rc=99;
	while(retries > 0 && rc != 0)
	{
		--retries;
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nMounting");
		sio_puts(string);
		delay_ms(50);	// Since mount will mess with the clock
#endif
    	rc=pf_mount(&fs); 		//Mount volume
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nMount returns %X", rc);
		sio_puts(string);
#endif
	}
	return rc;
}
//
// Read our config file. A straight textfile with data in the form of lines
// containing keys and values, ie:
//
// PROTO 0
// DEBUG 1
//

int Read_config()
{
int i,n,c,j, size;
WORD nbytes;
char key[PARMLEN];
char val[PARMLEN];

#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nReading config");
	sio_puts(string);
#endif
    MMC = pf_open(cfgfile); //Open the file
	rc = (int) MMC;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nOpen %s returns %X", cfgfile, MMC);
	sio_puts(string);
#endif
	if(rc != 0) return rc;
	//
	// OK, we seem to have a config file. Parse it to get values
	//
	rc=pf_read(buffer,256,&nbytes);
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nRead %d bytes", nbytes);
	sio_puts(string);
#endif
	//
	n=0;
	c=0;
	for(j=0;j<PARMLEN;j++)
	{
		key[j]=0;
		val[j]=0;
	}
	for(i=0;i<nbytes;i++)
	{
		//sprintf(string,(const far rom char*)"\r\n%d %x %d %d",i,buffer[i],c,n );
		//sio_puts(string);
		if(buffer[i] == 0x0d)
		{
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nK:%s V:%s", key, val);
		sio_puts(string);
		//
		if(strcmpram2pgm((const far rom char*)"PROTO", key)==0)
			protocol = atoi(val) & 0xff;
		if(!strcmpram2pgm((const far rom char*)"DEBUG", key))
			debuglevel = atoi(val) & 0xff;
		//	
		for(j=0;j<PARMLEN;j++)
		{
			key[j]=0;
			val[j]=0;
		}
#endif
			c=0;
			n=0;
		}
		else if(buffer[i] == 0x0a)
		{
		}
		else if((buffer[i]==' '))
		{
			c=1;
			n=0;
		}
		else
		{
			if(c) val[n] = buffer[i];
			else key[n] = buffer[i];
			if(n < PARMLEN) n++;
		}
	}
	sprintf(string,(const far rom char*)"\r\nP:%d D:%d", protocol, debuglevel);
	sio_puts(string);
	return 0;
}
//
// Blink the error code. Blink <code> short, followed by one long
//
void Error_blink(int code)
{
	LED1_PORT = 0; 
	delay_ms(1000);
	LED1_PORT = 1; 
	delay_ms(1000);
	while(code--)
	{
		LED1_PORT = 0; 
		delay_ms(300);
		LED1_PORT = 1;
		delay_ms(300);
	}
	LED1_PORT = 0; 
	delay_ms(500);
	LED1_PORT = 1; 
}
//
void Time(void)
{
unsigned int ticks = TMR0L + 256 * TMR0H;
	sprintf(string,(const far rom char*)"\r\nTicks:%u %d\r\n", ticks, INTCONbits.TMR0IF);
	sio_puts(string);
	INTCONbits.TMR0IF = 0;
}